from nashi import nav
from flask_security.core import current_user
from flask_nav.elements import Navbar, View, Subgroup, Link, Text, Separator


@nav.navigation()
def mynavbar():
    if current_user.is_anonymous:
        usergroup = Subgroup(
                    "User",
                    Text("Not logged in"),
                    View('Forgot password', 'security.forgot_password'),
                )
    else:
        usergroup = Subgroup(
                    "User",
                    Text(current_user.email),
                    View('Change password', 'security.change_password'),
                    Separator,
                    View('Logout', 'security.logout')
                )

    return Navbar(
        View('nashi', 'index'),
        Subgroup(
            'Links',
            Link('Philosophie Uni Würzburg', 'https://www.philosophie.' +
                 'uni-wuerzburg.de/forschung/forschungsstellephilosophie-un/'),
            Link('AL-Corpus', 'http://arabic-latin-corpus.philosophie.' +
                 'uni-wuerzburg.de/'),
            Link('AL-Corpus wiki', 'http://kallimachos.de/al-corpus'),
        ),
        usergroup
    )


@nav.navigation()
def indexnavbar():
    nav = mynavbar()
    nav.items.insert(0, Subgroup(
            'Library',
            View('Refresh list of books', 'index', action="refresh_booklist"),
            ))
    return nav



# -*- coding: utf-8 -*-

"""
Scans the LAREX_DIR for books and PageXML,
if book or page not in db or page.data empty:
create, load, pagesegment, update no_lines_segm
"""

# from nashi import app
from nashi.models import Book, Page
from nashi.database import db_session

import numpy as np
from skimage.measure import find_contours, approximate_polygon

from glob import glob
from os import path
from lxml import etree
from PIL import Image, ImageDraw

from kraken import pageseg, binarization
from kraken.lib import morph, sl
from kraken.lib.util import pil2array
from kraken.binarization import is_bitonal
from kraken.lib.exceptions import KrakenInputException


def cutout(im, coords):
    """
        Cut out coords from image, crop and return new image.
    """
    coords = [tuple(t) for t in coords]
    if not coords:
        return None
    maskim = Image.new('1', im.size, 0)
    ImageDraw.Draw(maskim).polygon(coords, outline=1, fill=1)
    new = Image.new(im.mode, im.size, "white")
    masked = Image.composite(im, new, maskim)
    cropped = masked.crop([
            min([x[0] for x in coords]), min([x[1] for x in coords]),
            max([x[0] for x in coords]), max([x[1] for x in coords])])
    return cropped


class record(object):
    def __init__(self, **kw):
        self.__dict__.update(kw)


def compute_lines(segmentation, spread, scale):
    """Given a line segmentation map, computes a list
    of tuples consisting of 2D slices and masked images."""
    lobjects = morph.find_objects(segmentation)
    lines = []
    for i, o in enumerate(lobjects):
        if o is None:
            continue
        if sl.dim1(o) < 2*scale or sl.dim0(o) < scale:
            continue
        mask = (segmentation[o] == i+1)
        if np.amax(mask) == 0:
            continue
        result = record()
        result.label = i+1
        result.bounds = o
        polygon = []
        if ((segmentation[o] != 0) == (segmentation[o] != i+1)).any():
            ppoints = draw_polygon(spread[o], i+1)
            ppoints = ppoints[1:] if ppoints else []
            polygon = [(o[0].start+p[0], o[1].start+p[1]) for p in ppoints]
        if not polygon:
            polygon = [(o[0].start, o[1].start), (o[0].stop,  o[1].start),
                       (o[0].stop,  o[1].stop),  (o[0].start, o[1].stop)]
        result.polygon = polygon
        result.mask = mask
        lines.append(result)
    return lines


def draw_polygon(lspread, lineno):
    """Draws a polygon around area of value lineno in array lspread."""
    lspread = np.pad(lspread, 1, "constant", constant_values=0)
    cont = find_contours(np.where(lspread == lineno, lineno, 2*lineno), lineno)
    if len(cont) == 1 and all(cont[0][0] == cont[0][-1]):
        polyg = approximate_polygon(cont[0], tolerance=1).astype(int)
        return [(p[0]-1, p[1]-1) for p in polyg]
    else:
        return []


def segment(im, text_direction='horizontal-lr', scale=None, maxcolseps=2,
            black_colseps=False):
    """
    Segments a page into text lines.
    Segments a page into text lines and returns the absolute coordinates of
    each line in reading order.
    Args:
        im (PIL.Image): A bi-level page of mode '1' or 'L'
        text_direction (str): Principal direction of the text
                              (horizontal-lr/rl/vertical-lr/rl)
        scale (float): Scale of the image
        maxcolseps (int): Maximum number of whitespace column separators
        black_colseps (bool): Whether column separators are assumed to be
                              vertical black lines or not
    Returns:
        {'text_direction': '$dir', 'boxes': [(x1, y1, x2, y2),...]}: A
        dictionary containing the text direction and a list of reading order
        sorted bounding boxes under the key 'boxes'.
    Raises:
        KrakenInputException if the input image is not binarized or the text
        direction is invalid.
    """

    if im.mode != '1' and not is_bitonal(im):
        raise KrakenInputException('Image is not bi-level')

    # rotate input image for vertical lines
    if text_direction.startswith('horizontal'):
        angle = 0
        offset = (0, 0)
    elif text_direction == 'vertical-lr':
        angle = 270
        offset = (0, im.size[1])
    elif text_direction == 'vertical-rl':
        angle = 90
        offset = (im.size[0], 0)
    else:
        raise KrakenInputException('Invalid text direction')

    im = im.rotate(angle, expand=True)

    # honestly I've got no idea what's going on here. In theory a simple
    # np.array(im, 'i') should suffice here but for some reason the
    # tostring/fromstring magic in pil2array alters the array in a way that is
    # needed for the algorithm to work correctly.
    a = pil2array(im)
    binary = np.array(a > 0.5*(np.amin(a) + np.amax(a)), 'i')
    binary = 1 - binary

    if not scale:
        scale = pageseg.estimate_scale(binary)

    binary = pageseg.remove_hlines(binary, scale)
    # emptyish images will cause exceptions here.
    try:
        if black_colseps:
            colseps, binary = pageseg.compute_black_colseps(binary, scale,
                                                            maxcolseps)
        else:
            colseps = pageseg.compute_white_colseps(binary, scale, maxcolseps)
    except ValueError:
        return {'text_direction': text_direction, 'boxes':  []}

    bottom, top, boxmap = pageseg.compute_gradmaps(binary, scale)
    seeds = pageseg.compute_line_seeds(binary, bottom, top, colseps, scale)
    llabels1 = morph.propagate_labels(boxmap, seeds, conflict=0)
    spread = morph.spread_labels(seeds, maxdist=scale)
    llabels = np.where(llabels1 > 0, llabels1, spread*binary)
    segmentation = llabels*binary

    lines_and_polygons = compute_lines(segmentation, spread, scale)
    # TODO: rotate_lines for polygons
    order = pageseg.reading_order([l.bounds for l in lines_and_polygons])
    lsort = pageseg.topsort(order)
    lines = [lines_and_polygons[i].bounds for i in lsort]
    lines = [(s2.start, s1.start, s2.stop, s1.stop) for s1, s2 in lines]
    return {'text_direction': text_direction,
            'boxes': pageseg.rotate_lines(lines, 360-angle, offset).tolist(),
            'lines': lines_and_polygons,
            'script_detection': False}


def pagexmllineseg(xmlfile, imgpath, text_direction='horizontal-lr'):
    root = etree.parse(xmlfile).getroot()
    ns = {"ns": root.nsmap[None]}

    # convert point notation from older pagexml versions
    for c in root.xpath("//ns:Coords[not(@points)]", namespaces=ns):
        cc = []
        for point in c.xpath("./ns:Point", namespaces=ns):
            # coordstrings = [x.split(",") for x in c.attrib["points"].split()]
            cx = point.attrib["x"]
            cy = point.attrib["y"]
            c.remove(point)
            cc.append(cx+","+cy)
        c.attrib["points"] = " ".join(cc)

    coordmap = {}
    for r in root.xpath('//ns:TextRegion', namespaces=ns):
        rid = r.attrib["id"]
        coordmap[rid] = {"type": r.attrib["type"]}
        coordmap[rid]["coords"] = []
        for c in r.xpath("./ns:Coords", namespaces=ns) + r.xpath("./Coords"):
            coordmap[rid]["coordstring"] = c.attrib["points"]
            coordstrings = [x.split(",") for x in c.attrib["points"].split()]
            coordmap[rid]["coords"] += [[int(x[0]), int(x[1])]
                                        for x in coordstrings]

    filename = root.xpath('//ns:Page', namespaces=ns)[0]\
        .attrib["imageFilename"]
    filename = imgpath + "/" + filename

    im = Image.open(filename)

    for n, c in enumerate(sorted(coordmap)):
        coords = coordmap[c]['coords']
        if len(coords) < 3:
            continue
        cropped = cutout(im, coords)
        offset = (min([x[0] for x in coords]), min([x[1] for x in coords]))
        if cropped is not None:
            if not binarization.is_bitonal(cropped):
                try:
                    cropped = binarization.nlbin(cropped)
                except SystemError:
                    continue
            if coordmap[c]["type"] == "drop-capital":
                lines = [1]
            else:
                # if line in
                lines = segment(cropped, text_direction=text_direction,
                                maxcolseps=-1)

                lines = lines["lines"] if "lines" in lines else []
        else:
            lines = []

        for n, l in enumerate(lines):
            if coordmap[c]["type"] == "drop-capital":
                coordstrg = coordmap[c]["coordstring"]
            else:
                coords = ((x[1]+offset[0], x[0]+offset[1]) for x in l.polygon)
                coordstrg = " ".join([str(x[0])+","+str(x[1]) for x in coords])
            textregion = root.xpath('//ns:TextRegion[@id="'+c+'"]',
                                    namespaces=ns)[0]
            linexml = etree.SubElement(textregion, "TextLine",
                                       attrib={"id": "{}_l{:03d}".format(
                                        c, n+1)})
            coordsxml = etree.SubElement(linexml, "Coords",
                                         attrib={"points": coordstrg})
    xmlstring = etree.tounicode(root.getroottree()).replace(
        "http://schema.primaresearch.org/PAGE/gts/pagecontent/2010-03-19",
        "http://schema.primaresearch.org/PAGE/gts/pagecontent/2017-07-15")
    no_lines_segm = int(root.xpath("count(//TextLine)"))
    return xmlstring, no_lines_segm


def create_book(bookpath):
    """ Return book database object for path in larex dir. """
    bookname = path.split(bookpath)[1]
    no_pages_total = len(glob(bookpath+"/*.png"))
    book = Book.query.filter_by(name=bookname).one()
    if not book:
        book = Book(name=bookname, no_pages_total=no_pages_total)
    else:
        book.no_pages_total = no_pages_total
    return book


def add_page(book, xmlfile, commit=True):
    """ Add page from PageXML file to book (segment lines, update version...).
        TODO: add reading order
    """
    text_direction = 'horizontal-rl'\
        if book.name.endswith("_ar") else 'horizontal-lr'
    bookpath = path.split(xmlfile)[0]
    pagename = path.split(xmlfile)[1].split(".")[0]
    page = Page.query.filter_by(book_id=book.id, name=pagename).one()
    if not page:
        page = Page(book=book, name=pagename)

    page.data, page.no_lines_segm = \
        pagexmllineseg(xmlfile, bookpath, text_direction=text_direction)
    if commit:
        db_session.add(book)
        db_session.commit()
    return page.no_lines_segm


if __name__ == '__main__':
    for bookpath in [x for x in glob(app.config["LAREX_DIR"]+"/*")
                     if path.isdir(x)]:

        book = create_book(bookpath)

        for xmlfile in sorted(glob(bookpath+"/*.xml")):
            add_page(book, xmlfile, commit=False)

        db_session.add(book)
        db_session.commit()



from nashi import app, make_celery
from nashi.import_from_larex import add_page
from nashi.models import Book
import time

celery = make_celery(app)


@celery.task(bind=True)
def lareximport(self, bookname, pages=[]):
    book = Book.query.filter_by(name=bookname).first()
    message = ''
    lines = 0
    for n, xmlfile in enumerate(pages):
        print(xmlfile)
        message = "Importing page {} of {}.".format(n+1, len(pages))
        self.update_state(state='PROGRESS',
                          meta={'current': n+1, 'total': len(pages),
                                'status': message})
        lines_added = add_page(book, xmlfile, commit=True)
        lines += lines_added
    return {'current': n+1, 'total': len(pages), 'status': 'Import completed!',
            'result': lines}



from flask import send_from_directory, render_template, make_response,\
    request, json, jsonify, flash, redirect, url_for, Response
from flask_security import login_required
from flask_security.core import current_user
import requests

import zipfile
import unicodedata
from lxml import etree, html
from glob import glob
from os import path
from io import BytesIO
from time import gmtime
from copy import deepcopy

from sqlalchemy.orm.exc import NoResultFound

from nashi import app
from nashi.models import Book, Page, EditorSettings
from nashi.database import db_session
from nashi.books import scan_bookfolder, copy_to_larex, upload_pagexml,\
    getlayers
from nashi.tasks import lareximport
from nashi.image import getsnippet


@app.route('/')
@login_required
def index():
    books = {}
    for book in Book.query.all():
        ocrstatus = book.ocrpid if book.ocrpid else "N/A"
        books[book.name] = {
                "no_pages_total": book.no_pages_total,
                "no_pages_segm": len(book.pages),
                "no_lines_segm": sum([p.no_lines_segm for p in book.pages]),
                "no_lines_gt": sum([p.no_lines_gt for p in book.pages]),
                "no_lines_ocr": sum([p.no_lines_ocr for p in book.pages]),
                "ocrstatus": ocrstatus
                }

    return render_template('index.html', books=books)


@app.route('/_library/<action>', methods=['POST', 'GET'])
@login_required
def library(action=""):
    print("action: " + action)
    if action == "refresh_booklist":
        scan_bookfolder(app.config["BOOKS_DIR"])
        res = jsonify(success=1)
    if action == "upload_pagexml":
        res = upload_pagexml(request.files["importzip"])
    return res


@app.route('/_libedit/<bookname>/<action>', methods=['POST', 'GET'])
@login_required
def libedit(bookname, action=""):
    book = Book.query.filter_by(name=bookname).first()
    if not book:
        return jsonify(success=0)

    if action == "delete":
        db_session.delete(book)
        db_session.commit()
        return jsonify(success=1)

    if action == "copy_to_larex":
        count = copy_to_larex(bookname, app.config["BOOKS_DIR"],
                              app.config["LAREX_DIR"], app.config["LAREX_GRP"])
        flash("Copied {} files to LAREX.".format(count))
        return jsonify(files_copied=count)

    if action == "select_from_larex":
        ps = glob("{}/{}/*.xml".format(app.config["LAREX_DIR"], bookname))
        pages = []
        for p in ps:
            pname = path.split(p)[1].split(".")[0]
            existing = Page.query.filter_by(
                book_id=book.id, name=pname).first()
            lines_ex = existing.no_lines_gt if existing else -1
            pages.append((pname, lines_ex))
        return jsonify(res=render_template('selectlarex.html', pages=pages))

    if action == "import_from_larex":
        data = request.json
        pages = data["pages"]
        pages = ["{}/{}/{}.xml".format(app.config["LAREX_DIR"], bookname, p)
                 for p in pages]
        for p in pages:
            if not path.isfile(p):
                pages[pages.index(p)] = p[:-3] + "bin.xml"
        task = lareximport.apply_async(
            args=[bookname], kwargs={"pages": pages})
        return jsonify({'Location': url_for('taskstatus', task_id=task.id)})


@app.route('/books/<bookname>/edit.html')
@login_required
def editor(bookname):
    book = Book.query.filter_by(name=bookname).first()
    if not book:
        return redirect(url_for("index"))
    pages = [(x.name, x.no_lines_gt, x.no_lines_segm) for x in book.pages]
    return render_template('editor.html', bookname=book.name, pages=pages)


@app.route('/_editorsettings', methods=['GET', 'POST'])
@login_required
def editorsettings():
    if current_user.is_anonymous:
        email = "user@nashi"
    else:
        email = current_user.email
    if request.method == "GET":
        try:
            s = EditorSettings.query.filter_by(email=email).one()
        except NoResultFound:
            return jsonify(status="fail")
        return jsonify(status="success", settings=json.loads(s.settings))

    if request.method == "POST":
        try:
            s = EditorSettings.query.filter_by(email=email).one()
        except NoResultFound:
            s = EditorSettings(email=email)
        s.settings = json.dumps(request.get_json())
        db_session.add(s)
        db_session.commit()
        return jsonify(status="success")


@app.route('/books/<bookname>/textedit.html')
@login_required
def textedit(bookname):
    return render_template("textedit.html", bookname=bookname)


@app.route('/books/<bookname>/textlayers.html', methods=['GET', 'POST'])
@login_required
def textlayers(bookname):
    data = request.get_json()
    b = Book.query.filter_by(name=bookname).one()

    if request.method == "GET":
        return jsonify(layers=getlayers(b))

    elif data["action"] == "copy":
        source = data["layer"]
        target = data["target"]
        ct = 0
        for p in Page.query.filter_by(book_id=b.id):
            root = etree.fromstring(p.data)
            ns = {"ns": root.nsmap[None]}
            for e in root.xpath('//ns:TextEquiv[@index="{}"]'.format(source),
                                namespaces=ns):
                tl = e.getparent()
                new = deepcopy(e)
                new.attrib["index"] = target
                old = e.xpath('../ns:TextEquiv[@index="{}"]'.format(target),
                              namespaces=ns)
                if old:
                    e.getparent().remove(old[0])
                e.getparent().append(new)
                ct += 1
            p.no_lines_gt = int(root.xpath('count(//ns:TextEquiv'
                                           '[@index="0"])', namespaces=ns))
            p.no_lines_ocr = int(root.xpath('count(//ns:TextLine'
                                            '[count(./ns:TextEquiv'
                                            '[@index>0])>0])',
                                            namespaces=ns))
            p.data = etree.tounicode(root.getroottree())
        db_session.commit()
        return jsonify(copied=ct)

    elif data["action"] == "delete":
        layer = int(data["layer"])
        ct = 0
        for p in Page.query.filter_by(book_id=b.id):
            root = etree.fromstring(p.data)
            ns = {"ns": root.nsmap[None]}
            for e in root.xpath('//ns:TextEquiv[@index="{}"]'.format(layer),
                                namespaces=ns):
                e.getparent().remove(e)
                ct += 1
            p.no_lines_gt = int(root.xpath('count(//ns:TextEquiv'
                                           '[@index="0"])', namespaces=ns))
            p.no_lines_ocr = int(root.xpath('count(//ns:TextLine'
                                            '[count(./ns:TextEquiv'
                                            '[@index>0])>0])',
                                            namespaces=ns))
            p.data = etree.tounicode(root.getroottree())
        db_session.commit()
        return jsonify(deleted=ct)


@app.route('/books/<bookname>/chartable.html', methods=['GET'])
@login_required
def chartable(bookname):
    layer = request.args.get("layer", "0")
    print("Layer: {}".format(layer))
    b = Book.query.filter_by(name=bookname).one()
    chars = ""
    for p in Page.query.filter_by(book_id=b.id):
        root = etree.fromstring(p.data)
        ns = {"ns": root.nsmap[None]}
        chars += "".join(root.xpath('//ns:TextLine/' +
                                    'ns:TextEquiv[@index="{}"]'.format(layer) +
                                    '/ns:Unicode/text()', namespaces=ns))
    chars = set(chars)
    table = []
    for c in chars:
        try:
            name = unicodedata.name(c)
        except ValueError:
            name = "unknown"
        table.append((c, 'U+{:04X}'.format(ord(c)), name))
    return render_template("chartable.html", bookname=bookname, results=table)


@app.route('/books/<bookname>/_textsearch.html', methods=['POST'])
@login_required
def textsearch(bookname):
    data = request.get_json()
    searchterm = data["searchterm"]
    layer = data["layer"]
    book = Book.query.filter_by(name=bookname).one()
    results = []

    for p in book.pages:
        root = etree.fromstring(p.data)
        ns = {"ns": root.nsmap[None]}
        found = root.xpath('//ns:TextEquiv[@index="{}"]'.format(layer) +
                           '/ns:Unicode/text()' +
                           '[contains(.,"{}")]'.format(searchterm),
                           namespaces=ns)
        for o in found:
            text = str(o)
            textregion = o.getparent().getparent().getparent()
            id = textregion.attrib["id"]
            if "comments" in textregion.attrib:
                comment = textregion.attrib["comments"]
            else:
                comment = ""
            results.append((p.name, id, text, comment))
        if len(results) > 10000:
            return render_template("_searchresults.html", results=results,
                                   cnt="{} (there could be more,".format(len(
                                    results)) + "max. exceeded)")
    return render_template("_searchresults.html", results=results,
                           cnt=len(results))


@app.route('/books/<bookname>/_textreplace.html', methods=['POST'])
@login_required
def textreplace(bookname):
    data = request.get_json()
    replacements = {}
    for r in data["replacements"]:
        if r["page"] not in replacements:
            replacements[r["page"]] = []
        replacements[r["page"]].append((r["line"], r["text"].strip()))
    layer = data["layer"]
    book = Book.query.filter_by(name=bookname).one()
    cnt = 0
    for pname, rs in replacements.items():
        p = Page.query.filter_by(book_id=book.id, name=pname).one()
        xml = p.data
        root = etree.fromstring(xml)
        ns = {"ns": root.nsmap[None]}
        for r in rs:
            uc = root.xpath('//ns:TextLine[@id="{}"]'.format(r[0]) +
                            '/ns:TextEquiv[@index="{}"]'.format(layer) +
                            '/ns:Unicode', namespaces=ns)[0]
            uc.text = r[1]
            cnt += 1
        p.data = etree.tounicode(root.getroottree())
    db_session.commit()
    return "Wrote {} lines to layer {}.".format(cnt, layer)


@app.route('/books/<bookname>/<pageno>/<lineid>.png')
@login_required
def getlineimage(bookname, pageno, lineid):
    book = Book.query.filter_by(name=bookname).one()
    xml = Page.query.filter_by(book_id=book.id, name=pageno).one().data
    root = etree.fromstring(xml)
    ns = {"ns": root.nsmap[None]}
    coords = root.xpath('//ns:TextLine[@id="{}"]/ns:Coords/@points'
                        .format(lineid), namespaces=ns)[0]
    fn = root.find(".//ns:Page", namespaces=ns).attrib["imageFilename"]
    if fn.endswith(".bin.png"):
        altfile = app.config['BOOKS_DIR']+"/"+bookname+"/"+fn[:-7]+"raw.png"
        if path.isfile(altfile):
            fn = fn[:-7] + "raw.png"
    im = getsnippet(app.config['BOOKS_DIR']+bookname+"/"+fn, coords)
    return Response(im, mimetype="image/png")


@app.route('/books/<bookname>_PageXML.zip')
@login_required
def getzip(bookname):
    book = Book.query.filter_by(name=bookname).one()
    zf = BytesIO()
    z = zipfile.ZipFile(zf, mode='w', compression=zipfile.ZIP_DEFLATED)
    for p in book.pages:
        z.writestr(bookname + "/" + p.name + ".xml", p.data)
        # with z.open(p.name + ".xml", "w") as cont:
        #    cont.write(p.data.encode("utf-8"))
    for contfile in z.filelist:
        contfile.date_time = gmtime()[:6]
    z.close()
    zf.seek(0)
    zipped = zf.read()
    zf.close()
    return Response(zipped, mimetype='application/zip')


@app.route('/books/<bookname>/<file>.xml')
@login_required
def getxml(bookname, file):
    book = Book.query.filter_by(name=bookname).first()
    xml = Page.query.filter_by(book_id=book.id, name=file).first().data
    return Response(xml, mimetype='text/xml')


@app.route('/books/<bookname>/<file>.png')
@login_required
def getpng(bookname, file):
    if file.endswith(".bin"):
        altfile = app.config['BOOKS_DIR']+"/"+bookname+"/"+file[:-3]+"raw.png"
        if path.isfile(altfile):
            file = file[:-3] + "raw"
    # return app.send_static_file("0003.png")
    return send_from_directory(app.config['BOOKS_DIR'],
                               bookname+"/"+file+".png")


@app.route('/books/<bookname>/<pagename>/search_continue', methods=['POST'])
@login_required
def search_continue(bookname, pagename):
    book = Book.query.filter_by(name=bookname).one()
    pnames = sorted([p.name for p in book.pages])
    data = request.json
    reverse = data["dir"] < 0
    searchterm = data["searchterm"]
    if reverse:
        pnames.reverse()
    pnames = pnames[pnames.index(pagename) + 1:]
    result = {"page": "", "line": ""}
    for p in pnames:
        page = Page.query.filter_by(book_id=book.id, name=p).one()
        root = etree.fromstring(page.data)
        ns = {"ns": root.nsmap[None]}
        found = root.xpath('//ns:TextLine', namespaces=ns)
        if reverse:
            found.reverse()
        for textline in found:
            indz = textline.xpath('./ns:TextEquiv/@index', namespaces=ns)
            lowestindex = min(indz) if indz else ""
            textcontent = textline.find(
                './ns:TextEquiv[@index="{}"]/ns:Unicode'.format(lowestindex),
                namespaces=ns).text if lowestindex else ""
            comm = ""
            if data["comments"] and "comments" in textline.attrib:
                comm = textline.attrib["comments"]
            if searchterm in textcontent or searchterm in comm:
                result["page"] = p
                result["line"] = textline.attrib["id"]
                break
        if result["page"]:
            break
    return jsonify(result=result)


@app.route('/books/<bookname>/<pagename>/data', methods=['POST', 'GET'])
@login_required
def pagedata(bookname, pagename):
    pnamesplits = pagename.split("::")
    command = ""
    if len(pnamesplits) == 2:
        pagename, command = pnamesplits
    book = Book.query.filter_by(name=bookname).first()
    if command:
        plist = sorted([x.name for x in
                        Book.query.filter_by(name=bookname).first().pages])
        if command == "first":
            pagename = min(plist)
        elif command == "next":
            pagename = plist[(plist.index(pagename) + 1) % len(plist)]
        elif command == "prev":
            pagename = plist[(plist.index(pagename) - 1) % len(plist)]

    page = Page.query.filter_by(book_id=book.id, name=pagename).first()
    root = etree.fromstring(page.data)
    ns = {"ns": root.nsmap[None]}

    if request.method == "GET":
        if (request.args.get('download', '', type=str) == 'xml'):
            response = make_response(page.data)
            response.headers['Cache-Control'] = 'no-cache'
            response.headers['Content-Type'] = 'text/xml'
            response.headers['Content-Disposition'] =\
                "attachment; filename={}_{}.xml".format(bookname, pagename)
            return response

        pageattr = root.find(".//ns:Page", namespaces=ns).attrib
        image = {"file": pageattr["imageFilename"],
                 "image_x": pageattr["imageWidth"],
                 "image_y": pageattr["imageHeight"]}
        direction = "rtl" if bookname.endswith("_ar") else "ltr"
        regionmap = {}
        for r in root.findall(".//ns:TextRegion/ns:Coords[@points]",
                              namespaces=ns):
            textregion = r.getparent()
            r_id = textregion.attrib["id"]
            regionmap[r_id] = {}
            regionmap[r_id]["points"] = r.attrib["points"]
        pagemap = {}
        for l in root.findall(".//ns:TextLine/ns:Coords[@points]",
                              namespaces=ns):
            textline = l.getparent()
            l_id = textline.attrib["id"]
            pagemap[l_id] = {}
            pagemap[l_id]["points"] = l.attrib["points"]
            pagemap[l_id]["region"] = textline.getparent().attrib["id"]
            pagemap[l_id]["comments"] = textline.attrib["comments"] if \
                "comments" in textline.attrib else ""
            indz = textline.xpath('./ns:TextEquiv/@index', namespaces=ns)
            lowestindex = min(indz) if indz else ""
            textcontent = textline.find(
                './ns:TextEquiv[@index="{}"]/ns:Unicode'.format(lowestindex),
                namespaces=ns).text if lowestindex else ""
            status = "empty"
            if textcontent:
                status = "ocr" if int(lowestindex) else "gt"
            pagemap[l_id]["text"] = {"status": status, "content": textcontent}
        return jsonify(page=pagename, image=image, lines=pagemap,
                       regions=regionmap, direction=direction)

    if request.method == "POST":
        data = request.json
        if current_user.is_anonymous:
            user = "user@nashi"
        else:
            user = current_user.email

        for l in [l for l in data["edits"] if l["action"] == "delete"]:
            cur = l["id"]
            ldata = l
            line = root.find('.//ns:TextLine[@id="'+cur+'"]', namespaces=ns)
            line.getparent().remove(line)

        for l in [l for l in data["edits"] if l["action"] == "create"]:
            cur = l["id"]
            ldata = l
            region = root.find('.//ns:TextRegion[@id="{}"]'.format(
                ldata["input"]["region"]), namespaces=ns)
            line = etree.SubElement(region, "{{{}}}TextLine".format(
                ns["ns"]), attrib={"id": cur})
            coords = etree.SubElement(line, "{{{}}}Coords".format(ns["ns"]),
                                      attrib={"points":
                                              ldata["input"]["points"]})

        for l in [l for l in data["edits"] if l["action"] == "change"]:
            cur = l["id"]
            ldata = l
            text = ldata["input"]["text"]["content"].strip()
            textstatus = ldata["input"]["text"]["status"]
            comments = ldata["input"]["comments"]
            points = ldata["input"]["points"]
            rid = ldata["input"]["region"]
            line = root.find('.//ns:TextLine[@id="'+cur+'"]',
                             namespaces=ns)
            line.attrib["comments"] = comments
            if textstatus == "edit":
                tequiv = line.find('.//ns:TextEquiv[@index="0"]',
                                   namespaces=ns)
                if tequiv is None:
                    tequiv = etree.SubElement(line, "TextEquiv",
                                              attrib={"index": "0"})
                    unicodexml = etree.SubElement(tequiv, "Unicode")
                else:
                    unicodexml = tequiv.find('./ns:Unicode', namespaces=ns)
                tequiv.attrib["comments"] = "User: " + user
                unicodexml.text = text
            else:  # points changed
                coords = line.find('.//ns:Coords', namespaces=ns)
                coords.attrib["points"] = points

        page.no_lines_gt = int(root.xpath('count(//TextEquiv[@index="0"])')) +\
            int(root.xpath('count(//ns:TextEquiv[@index="0"])', namespaces=ns))
        page.no_lines_segm = int(root.xpath('count(//TextLine)')) + \
            int(root.xpath('count(//ns:TextLine)', namespaces=ns))
        page.data = etree.tounicode(root.getroottree())
        db_session.commit()
        return jsonify(lineinfo=str(page.no_lines_gt)
                       + "/" + str(page.no_lines_segm))


@app.route('/bootstrapexample')
@login_required
def bootstrapexample():
    return render_template('bootstrapexample.html')


@app.route('/larexredir/<bookname>')
@login_required
def larexredir(bookname):
    if not(len(glob(app.config["LAREX_DIR"]+"/"+bookname))):
        flash("Book not in LAREX folder. Copy images first.")
        return redirect(url_for("index"))
    try:
        page = requests.get(app.config['LAREX_URL_SERVER'])
    except requests.exceptions.ConnectionError:
        flash("No connection to LAREX server!")
        return redirect(url_for("index"))
    html_content = html.fromstring(page.content)
    ids = html_content.xpath('.//div[@id="lib"]//tr[td = "'+bookname+'"]/@id')
    if len(ids) == 1:
        return redirect(app.config['LAREX_URL_CLIENT']+"viewer?book="+ids[0])
    else:
        flash("Book not found in LAREX. Check your server configuration.")
        return redirect(url_for("index"))


@app.route('/status/<task_id>')
@login_required
def taskstatus(task_id):
    task = lareximport.AsyncResult(task_id)
    if task.state == 'PENDING':
        # job did not start yet
        response = {
            'state': task.state,
            'current': 0,
            'total': 1,
            'status': 'Pending...'
        }
    elif task.state != 'FAILURE':
        response = {
            'state': task.state,
            'current': task.info.get('current', 0),
            'total': task.info.get('total', 1),
            'status': task.info.get('status', '')
        }
        if 'result' in task.info:
            response['result'] = task.info['result']
    else:
        # something went wrong in the background job
        response = {
            'state': task.state,
            'current': 1,
            'total': 1,
            'status': str(task.info),  # this is the exception raised
        }
    return jsonify(response)


@app.route('/loggedin')
@login_required
def loggedin():
    # authentification for nginx
    return ""



# -*- coding: utf-8 -*-

import numpy as np
import skimage.io as skimage_io
from skimage.draw import polygon
from io import BytesIO
from PIL import Image


def cutout(pageimg, coordstring, scale=1):
    """Cuts out coords from pageimg."""
    coords = [p.split(",") for p in coordstring.split()]
    coords = np.array([(int(scale*int(c[1])), int(scale*int(c[0])))
                      for c in coords])
    rr, cc = polygon(coords[:, 0], coords[:, 1], pageimg.shape)
    offset = (min([x[0] for x in coords]), min([x[1] for x in coords]))
    box = np.ones(
        (max([x[0] for x in coords]) - offset[0],
         max([x[1] for x in coords]) - offset[1]),
        dtype=pageimg.dtype) * 255
    box[rr-offset[0], cc-offset[1]] = pageimg[rr, cc]
    return box, offset


def getsnippet(filename, coordstring):
    cut = cutout(skimage_io.imread(filename), coordstring)[0]
    file = BytesIO()
    image = Image.fromarray(cut)
    image.save(file, 'png')
    file.name = "line.png"
    file.seek(0)
    return file.read()
